from __future__ import annotations

__doc__ = """
.. autofunction:: canonicalize_einsum
"""

__copyright__ = "Copyright (C) 2022 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import abc
from dataclasses import dataclass
from functools import cached_property, reduce
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from collections.abc import Mapping

import numpy as np
from bidict import frozenbidict
from immutables import Map
from pytools import UniqueNameGenerator, memoize_method

from feinsum.einsum import (
    INT_CLASSES,
    BatchedEinsum,
    IntegralT,
    ShapeComponentT,
    SizeParam,
)

# {{{ private interface to build an einsum graph


class EinsumGraphNode(abc.ABC):
    """
    A node in the einsum graph to canonicalize it.
    """

    @property
    @abc.abstractmethod
    def label(self) -> str:
        """
        Label used for graphviz visualization.
        """
        pass


@dataclass(frozen=True)
class ArgNode(EinsumGraphNode):
    name: str

    @property
    def label(self) -> str:
        return f"{self.name}"


@dataclass(frozen=True)
class IndexNode(EinsumGraphNode):
    name: str

    @property
    def label(self) -> str:
        return f"{self.name}"


@dataclass(frozen=True)
class AxisLengthNode(EinsumGraphNode):
    length: ShapeComponentT

    @property
    def label(self) -> str:
        name = (
            self.length.name
            if isinstance(self.length, SizeParam)
            else str(self.length)
        )
        return f"Length:{name}"


@dataclass(frozen=True)
class DimNode(EinsumGraphNode):
    dim: int

    @property
    def label(self) -> str:
        return f"dim-{self.dim + 1}"


@dataclass(frozen=True)
class DtypeNode(EinsumGraphNode):
    dtype: np.dtype[Any]

    @property
    def label(self) -> str:
        return f"{self.dtype}"


@dataclass(frozen=True)
class InputAccessNode(EinsumGraphNode):
    i: int
    j: int
    index: str
    d: int

    @property
    def label(self) -> str:
        return f"({self.i + 1}, {self.j + 1}, {self.index}, {self.d + 1})"


@dataclass(frozen=True)
class OutputAccessNode(EinsumGraphNode):
    index: str
    d: int

    @property
    def label(self) -> str:
        return f"({self.index}, {self.d + 1})"


@dataclass(frozen=True)
class IResultNode(EinsumGraphNode):
    i: int

    @property
    def label(self) -> str:
        return f"R-{self.i + 1}"


@dataclass(frozen=True)
class IPositionNode(EinsumGraphNode):
    j: int

    @property
    def label(self) -> str:
        return f"Pos-{self.j + 1}"


def to_bliss_color(v: EinsumGraphNode) -> int:
    if isinstance(v, ArgNode):
        return 1
    elif isinstance(v, IndexNode):
        return 2
    elif isinstance(v, InputAccessNode):
        return 3
    elif isinstance(v, OutputAccessNode):
        return 4
    elif isinstance(v, IResultNode):
        return 5
    elif isinstance(v, IPositionNode):
        return 6
    elif isinstance(v, DtypeNode):
        return 7
    elif isinstance(v, AxisLengthNode):
        return 8
    elif isinstance(v, DimNode):
        return 9
    else:
        raise NotImplementedError(f"Unknown node type '{type(v)}'.")


def to_graphviz_color(v: EinsumGraphNode) -> str:
    if isinstance(v, ArgNode):
        return "deepskyblue"
    elif isinstance(v, IndexNode):
        return "dodgerblue4"
    elif isinstance(v, InputAccessNode):
        return "darkseagreen1"
    elif isinstance(v, OutputAccessNode):
        return "forestgreen"
    elif isinstance(v, IResultNode):
        return "darksalmon"
    elif isinstance(v, IPositionNode):
        return "crimson"
    elif isinstance(v, DtypeNode):
        return "navajowhite"
    elif isinstance(v, AxisLengthNode):
        return "orange"
    elif isinstance(v, DimNode):
        return "thistle"
    else:
        raise NotImplementedError(f"Unknown node type '{type(v)}'.")


# }}}


def _axis_len_le(len1: ShapeComponentT, len2: ShapeComponentT) -> bool:
    """
    Returns true only if len1 < len2.
    """
    if isinstance(len1, SizeParam) and not isinstance(len2, SizeParam):
        return False
    elif isinstance(len2, SizeParam) and not isinstance(len1, SizeParam):
        return True
    elif isinstance(len1, SizeParam) and isinstance(len2, SizeParam):
        return False
    else:
        assert isinstance(len1, INT_CLASSES)
        assert isinstance(len2, INT_CLASSES)
        return bool(len1 < len2)


def visualize_einsum_graph(
    graph: Mapping[EinsumGraphNode, frozenset[EinsumGraphNode]],
) -> None:
    """
    Opens an Xwindow with the einsum graph connectivity
    as visualized by DOT.
    """
    from pytools.graphviz import show_dot

    node_to_name: dict[EinsumGraphNode, str] = {}
    vng = UniqueNameGenerator({"node"})

    all_nodes = frozenset(graph)
    assert all(val <= all_nodes for val in graph.values())
    graphviz_node_lines: list[str] = []
    graphviz_edge_lines: list[str] = []

    for node in all_nodes:
        name = vng("node")
        node_to_name[node] = name
        graphviz_node_lines.append(
            f'  {name}[label="{node.label}", color={to_graphviz_color(node)}];'
        )

    for node, successors in graph.items():
        pred = node_to_name[node]
        for successor in successors:
            succ = node_to_name[successor]
            graphviz_edge_lines.append(f"  {pred} -> {succ};")

    dot_lines = [
        "digraph {",
        "node[style=filled]",
        *graphviz_node_lines,
        "",
        "",
        *graphviz_edge_lines,
        "}",
    ]
    dot_code = "\n".join(dot_lines)

    show_dot(dot_code)


@dataclass(frozen=True)
class InducedDirectedGraph:
    edges: np.ndarray[tuple[int, int], np.dtype[np.uint64]]
    c: np.ndarray[tuple[int], np.dtype[np.uint8]]
    iota_dtype: Map[int, np.dtype[Any]]
    iota_length: Map[int, IntegralT]

    def __post_init__(self) -> None:
        assert self.edges.ndim == 2
        assert self.c.ndim == 1
        assert self.edges.shape[1] == 2
        n_idg = self.c.shape[0]
        assert all(0 <= n < n_idg for n in self.iota_dtype)
        assert all(0 <= n < n_idg for n in self.iota_length)
        assert len(self.iota_dtype) == len(set(self.iota_dtype.values()))
        assert len(self.iota_length) == len(set(self.iota_length.values()))

    @cached_property
    def _preds_and_succs(
        self,
    ) -> tuple[Mapping[int, frozenset[int]], Mapping[int, frozenset[int]]]:
        # TODO: Switch to numpy / scipy.sparse if dealing with huge graphs.
        preds: dict[int, set[int]] = {i: set() for i in range(self.n_idg)}
        succs: dict[int, set[int]] = {i: set() for i in range(self.n_idg)}
        for from_node, to_node in self.edges:
            succs[from_node].add(to_node)
            preds[to_node].add(from_node)

        frozen_preds = Map((k, frozenset(v)) for k, v in preds.items())
        frozen_succs = Map((k, frozenset(v)) for k, v in succs.items())

        return frozen_preds, frozen_succs

    @property
    def n_idg(self) -> int:
        return self.c.shape[0]

    @memoize_method
    def preds(self, i: int) -> frozenset[int]:
        """
        Returns all the predecessors of *i*-th node.
        """
        preds, _ = self._preds_and_succs
        return preds[i]

    @memoize_method
    def succs(self, i: int) -> frozenset[int]:
        """
        Returns all the successors of *i*-th node.
        """
        _, succs = self._preds_and_succs
        return succs[i]

    def visualize(self, output_to: str | None = None) -> None:
        graphviz_colors = {
            1: "deepskyblue",
            2: "dodgerblue4",
            3: "darkseagreen1",
            4: "forestgreen",
            5: "darksalmon",
            6: "crimson",
            7: "navajowhite",
            8: "orange",
            9: "thistle",
        }

        dot_code = "digraph {\n"
        dot_code += "node[style=filled]\n"
        for i, color in enumerate(self.c):
            dot_code += f"{i} [color={graphviz_colors[int(color)]}]\n"
        for i, j in self.edges:
            dot_code += f"{i} -> {j}\n"
        dot_code += "}"

        from pytools.graphviz import show_dot

        show_dot(dot_code, output_to)

    @memoize_method
    def to_bliss_canonicalized_induced_dag(
        self,
    ) -> tuple[InducedDirectedGraph, np.ndarray[tuple[int], np.dtype[np.uint64]]]:
        import pybliss as bliss

        self_digraph = bliss.digraph_from_numpy(self.n_idg, self.edges, self.c)
        bliss_stats = bliss.Stats()
        perm = self_digraph.get_permutation_to_canonical_form(bliss_stats)
        assert isinstance(perm, np.ndarray)

        new_edges, new_colors = perm[self.edges], self.c[perm]
        new_iota_dtype = Map(
            (perm[i], dtype) for i, dtype in self.iota_dtype.items()
        )
        new_iota_length = Map(
            (perm[i], dtype) for i, dtype in self.iota_length.items()
        )
        return (
            InducedDirectedGraph(
                new_edges, new_colors, new_iota_dtype, new_iota_length
            ),
            perm,
        )


def to_induced_dag(
    einsum: BatchedEinsum,
) -> tuple[
    InducedDirectedGraph,
    Mapping[int, str],
]:
    """
    Returns the induced directed graph for a batched einsum *einsum*.
    """
    # Implementation follows Appendix 'A' of <PAPER NAME>.

    # See Defn 16.
    n_dim = max([einsum.ndim, *[len(idx_set) for idx_set in einsum.in_idx_sets]])
    all_dims = frozenset(DimNode(i) for i in range(n_dim))

    # See Defn. 17
    input_accesses = frozenset(
        {
            InputAccessNode(i, j, idx, d)
            for i in range(einsum.b)
            for j, idx_set in enumerate(einsum.in_idx_sets)
            for d, idx in enumerate(idx_set)
        }
    )

    # See Defn. 18
    output_accesses = frozenset(
        {OutputAccessNode(idx, d) for d, idx in enumerate(einsum.out_idx_set)}
    )

    # See Defn 19
    dtype_nodes = frozenset(
        {DtypeNode(dtype) for dtype in einsum.arg_to_dtype.values()}
    )

    # See Defn 20
    axis_lengths = frozenset(
        {
            AxisLengthNode(axis_len)
            for axis_len in einsum.index_to_dim_length.values()
        }
    )

    # See Defn 9
    all_indices = frozenset({IndexNode(idx) for idx in einsum.all_indices})

    # See Defn 12
    all_args = frozenset({ArgNode(arg_name) for arg_name in einsum.all_args})

    n_arg = len(all_args)
    n_index = len(all_indices)
    n_length = len(axis_lengths)
    n_dtypes = len(dtype_nodes)
    n_access_in = len(input_accesses)
    n_access_out = len(output_accesses)

    # {{{ See Section A.1, define the iota mappings

    iota_arg = frozenbidict(
        (i, arg_node)
        for i, arg_node in enumerate(
            sorted(all_args, key=lambda arg_node: arg_node.name),
        )
    )
    iota_index = frozenbidict(
        (i, index_node)
        for i, index_node in enumerate(
            sorted(all_indices, key=lambda index_node: index_node.name),
            start=n_arg,
        )
    )

    iota_access_in = frozenbidict(
        (i, access_in_node)
        for i, access_in_node in enumerate(
            sorted(
                input_accesses,
                key=lambda access_in_node: (
                    access_in_node.i,
                    access_in_node.j,
                    access_in_node.d,
                    access_in_node.index,
                ),
            ),
            start=n_arg + n_index,
        )
    )

    iota_access_out = frozenbidict(
        (i, access_out_node)
        for i, access_out_node in enumerate(
            sorted(
                output_accesses,
                key=lambda access_out_node: (
                    access_out_node.d,
                    access_out_node.index,
                ),
            ),
            start=n_arg + n_index + n_access_in,
        )
    )

    iota_output = frozenbidict(
        (
            (i_result + n_arg + n_index + n_access_in + n_access_out),
            IResultNode(i_result),
        )
        for i_result in range(einsum.b)
    )

    iota_pos = frozenbidict(
        (
            (i_pos + n_arg + n_index + n_access_in + n_access_out + einsum.b),
            IPositionNode(i_pos),
        )
        for i_pos in range(einsum.n)
    )

    iota_dtype = frozenbidict(
        (i, dtype_node)
        for i, dtype_node in enumerate(
            sorted(
                dtype_nodes,
                key=lambda dtype_node: str(dtype_node.dtype),
            ),
            start=n_arg + n_index + n_access_in + n_access_out + einsum.b + einsum.n,
        )
    )

    iota_length = frozenbidict(
        (i, length_node)
        for i, length_node in enumerate(
            sorted(
                axis_lengths,
                key=lambda length_node: str(length_node),
            ),
            start=n_arg
            + n_index
            + n_access_in
            + n_access_out
            + einsum.b
            + einsum.n
            + n_dtypes,
        )
    )

    iota_dim = frozenbidict(
        (i, dim_node)
        for i, dim_node in enumerate(
            sorted(
                all_dims,
                key=lambda dim_node: str(dim_node),
            ),
            start=n_arg
            + n_index
            + n_access_in
            + n_access_out
            + einsum.b
            + einsum.n
            + n_dtypes
            + n_length,
        )
    )

    iotas: frozenbidict[int, EinsumGraphNode] = frozenbidict(
        {
            **iota_arg,
            **iota_index,
            **iota_access_in,
            **iota_access_out,
            **iota_output,
            **iota_pos,
            **iota_dtype,
            **iota_length,
            **iota_dim,
        }
    )
    assert (
        len(iotas)
        == n_arg
        + n_index
        + n_access_in
        + n_access_out
        + einsum.b
        + einsum.n
        + n_dtypes
        + n_length
        + n_dim
    )

    # }}}

    # {{{ record the edges (See (7) -- (17))

    e_access_in_to_arg = frozenset(
        {
            (InputAccessNode(i, j, idx, i_idx), ArgNode(arg.name))
            for i in range(einsum.b)
            for j, (idx_set, arg) in enumerate(
                zip(einsum.in_idx_sets, einsum.args[i], strict=True)
            )
            for i_idx, idx in enumerate(idx_set)
        }
    )

    e_arg_pos_to_access_in = frozenset(
        {
            (IPositionNode(access_in_node.j), access_in_node)
            for access_in_node in input_accesses
        }
    )
    e_out_to_access_in = frozenset(
        {
            (IResultNode(access_in_node.i), access_in_node)
            for access_in_node in input_accesses
        }
    )
    e_idx_to_access_in = frozenset(
        {
            (IndexNode(access_in_node.index), access_in_node)
            for access_in_node in input_accesses
        }
    )
    e_dim_to_access_in = frozenset(
        {
            (DimNode(access_in_node.d), access_in_node)
            for access_in_node in input_accesses
        }
    )
    e_idx_to_access_out = frozenset(
        {(IndexNode(access_out.index), access_out) for access_out in output_accesses}
    )
    e_dim_to_access_out = frozenset(
        {(DimNode(access_out.d), access_out) for access_out in output_accesses}
    )

    e_length = frozenset(
        {
            (AxisLengthNode(axis_len), IndexNode(idx))
            for idx, axis_len in einsum.index_to_dim_length.items()
        }
    )
    e_dtype = frozenset(
        {
            (DtypeNode(dtype), ArgNode(arg_name))
            for arg_name, dtype in einsum.arg_to_dtype.items()
        }
    )

    e_length_ranks_tmp = set()
    for length_i in axis_lengths:
        for length_j in axis_lengths:
            if _axis_len_le(length_i.length, length_j.length):
                e_length_ranks_tmp.add((length_i, length_j))
    e_length_ranks = frozenset(e_length_ranks_tmp)
    del e_length_ranks_tmp

    e_dtype_ranks_tmp = set()
    for dtype_i in dtype_nodes:
        for dtype_j in dtype_nodes:
            if str(dtype_i.dtype) < str(dtype_j.dtype):
                e_dtype_ranks_tmp.add((dtype_i, dtype_j))
    e_dtype_ranks = frozenset(e_dtype_ranks_tmp)
    del e_dtype_ranks_tmp

    e_dim_ranks_tmp = set()
    for dim_i in all_dims:
        for dim_j in all_dims:
            if dim_i.dim < dim_j.dim:
                e_dim_ranks_tmp.add((dim_i, dim_j))
    e_dim_ranks = frozenset(e_dim_ranks_tmp)
    del e_dim_ranks_tmp

    all_edges: frozenset[tuple[EinsumGraphNode, EinsumGraphNode]] = frozenset(
        {
            *e_access_in_to_arg,
            *e_arg_pos_to_access_in,
            *e_out_to_access_in,
            *e_idx_to_access_in,
            *e_arg_pos_to_access_in,
            *e_dim_to_access_in,
            *e_idx_to_access_out,
            *e_dim_to_access_out,
            *e_length,
            *e_dtype,
            *e_length_ranks,
            *e_dtype_ranks,
            *e_dim_ranks,
        }
    )

    # }}}

    if 0:
        graphviz_graph_tmp: dict[EinsumGraphNode, set[EinsumGraphNode]] = {
            node: set() for node in iotas.values()
        }
        for from_node, to_node in all_edges:
            graphviz_graph_tmp[from_node].add(to_node)
        graphviz_graph = Map(
            {
                from_node: frozenset(to_nodes)
                for from_node, to_nodes in graphviz_graph_tmp.items()
            }
        )
        del graphviz_graph_tmp

        visualize_einsum_graph(graphviz_graph)
        _ = 1 / 0

    # {{{ Instantiate the induced directed graph

    n_idg = len(iotas)
    bliss_colors = np.empty(n_idg, np.uint8)

    for i, node in iotas.items():
        bliss_colors[i] = to_bliss_color(node)

    bliss_edges = np.array(
        [
            [iotas.inv[from_node], iotas.inv[to_node]]
            for (from_node, to_node) in all_edges
        ],
        dtype=np.uint64,
    )

    # }}}

    return (
        InducedDirectedGraph(
            bliss_edges,
            bliss_colors,
            Map((i, dtype_node.dtype) for i, dtype_node in iota_dtype.items()),
            Map(
                (i, axis_len.length)
                for i, axis_len in iota_length.items()
                if isinstance(axis_len.length, INT_CLASSES)
            ),
        ),
        Map(
            {
                **{i: arg_node.name for i, arg_node in iota_arg.items()},
                **{i: index_node.name for i, index_node in iota_index.items()},
                **{
                    i: axis_len.length.name
                    for i, axis_len in iota_length.items()
                    if isinstance(axis_len.length, SizeParam)
                },
                **{
                    i: "_fe_out" if i_output < 1 else f"_fe_out_{i_output - 1}"
                    for i_output, i in enumerate(iota_output)
                },
            }
        ),
    )


def from_induced_dag(
    induced_dag: InducedDirectedGraph,
) -> tuple[BatchedEinsum, Mapping[int, str]]:
    # Follows the reconstruction from Appendix A.2

    # Equation (20)
    v_arg = frozenset(np.nonzero(induced_dag.c == 1)[0])
    v_index = frozenset(np.nonzero(induced_dag.c == 2)[0])
    v_access_in = frozenset(np.nonzero(induced_dag.c == 3)[0])
    v_access_out = frozenset(np.nonzero(induced_dag.c == 4)[0])
    v_output = frozenset(np.nonzero(induced_dag.c == 5)[0])
    v_arg_pos = frozenset(np.nonzero(induced_dag.c == 6)[0])
    v_dtype = frozenset(np.nonzero(induced_dag.c == 7)[0])
    v_length = frozenset(np.nonzero(induced_dag.c == 8)[0])
    v_dims = frozenset(np.nonzero(induced_dag.c == 9)[0])

    # {{{ Check the constaints (C1) -- (C30)

    # C1.
    assert np.all(np.logical_and(induced_dag.c >= 1, induced_dag.c <= 9))
    # C2.
    assert all(len(induced_dag.succs(i)) == 0 for i in v_arg)
    # C3.
    v_access_in_union_dtype = v_access_in | v_dtype
    assert all(induced_dag.preds(i) <= v_access_in_union_dtype for i in v_arg)
    # C4.
    assert all(len(induced_dag.preds(i) & v_dtype) == 1 for i in v_arg)
    # C5.
    assert all(induced_dag.preds(i) <= v_dtype for i in v_dtype)
    # C6.
    v_arg_u_v_dtype = v_arg | v_dtype
    assert all(induced_dag.succs(i) <= v_arg_u_v_dtype for i in v_dtype)
    # C7.
    assert all(induced_dag.succs(i) <= v_arg for i in v_access_in)
    # C8.
    assert all(
        (len(induced_dag.preds(i) & v_arg_pos) == 1)
        and (len(induced_dag.preds(i) & v_index) == 1)
        and (len(induced_dag.preds(i) & v_output) == 1)
        and (len(induced_dag.preds(i) & v_dims) == 1)
        and (len(induced_dag.preds(i)) == 4)
        for i in v_access_in
    )
    # C9.
    assert all(induced_dag.succs(i) <= v_access_in for i in v_output)
    # C10.
    assert all(len(induced_dag.preds(i)) == 0 for i in v_output)
    # C11.
    assert all(len(induced_dag.succs(i)) == 0 for i in v_access_out)
    # C12.
    assert all(
        (len(induced_dag.preds(i) & v_index) == 1)
        and (len(induced_dag.preds(i) & v_dims) == 1)
        and (len(induced_dag.preds(i)) == 2)
        for i in v_access_out
    )
    # C13.
    v_access_in_union_out = v_access_in | v_access_out
    assert all(induced_dag.succs(i) <= v_access_in_union_out for i in v_index)

    # C14.
    assert all(
        (len(induced_dag.preds(i) & v_length) == 1)
        and (len(induced_dag.preds(i)) == 1)
        for i in v_index
    )

    # C15.
    assert all(induced_dag.preds(i) <= v_length for i in v_length)
    # C16.
    v_index_union_length = v_index | v_length
    assert all(induced_dag.succs(i) <= v_index_union_length for i in v_length)

    # C17.
    assert all(induced_dag.preds(i) <= v_dims for i in v_dims)

    # C18.
    v_access_in_union_out_union_dim = v_access_in | v_access_out | v_dims
    assert all(
        induced_dag.succs(i) <= v_access_in_union_out_union_dim for i in v_dims
    )

    # C19.
    assert all(len(induced_dag.preds(i)) == 0 for i in v_arg_pos)

    # C20.
    assert all(induced_dag.succs(i) <= v_access_in for i in v_arg_pos)

    # C21. (FIXME). It is incorrect in the paper.

    # C22. (FIXME). It is incorrect in the paper.

    # C23. (FIXME). It is incorrect in the paper.

    # C24. (FIXME). Not sure what the paper says.

    # C25.
    for n in v_access_out:
        (i,) = induced_dag.preds(n) & v_index
        assert len(induced_dag.succs(i) & v_access_in) > 0

    # C26. (FIXME)

    # C27. (FIXME)

    # C28. (FIXME)

    # C29. (FIXME)

    # C30. (FIXME)

    # }}}

    # Step 1.
    def arg_name_gen(i: int) -> str:
        return f"arg_{i}"

    def idx_name_gen(i: int) -> str:
        assert i < 26
        return chr((ord("i") - ord("a") + i) % 26 + ord("a"))

    # Step 2.
    iota_inferred_arg_pos = Map(
        (arg_pos, i) for i, arg_pos in enumerate(sorted(v_arg_pos))
    )

    # Step 3.
    iota_inferred_output = Map(
        (output, i) for i, output in enumerate(sorted(v_output))
    )

    # Step 4.
    iota_inferred_index = Map(
        (i, idx_name_gen(i_idx)) for i_idx, i in enumerate(sorted(v_index))
    )

    # Step 5.
    iota_inferred_arg = Map(
        (i, arg_name_gen(i_arg)) for i_arg, i in enumerate(sorted(v_arg))
    )

    # Step 6.
    iota_inferred_dim = Map(
        (idim, len(induced_dag.preds(idim) & v_dims)) for idim in v_dims
    )
    assert sorted(iota_inferred_dim.values()) == list(range(len(v_dims)))

    # Step 7.
    i_inferred_dim_to_out_index_tmp = {}
    for i in v_access_out:
        (i_dim,) = induced_dag.preds(i) & v_dims
        (i_idx,) = induced_dag.preds(i) & v_index
        i_inferred_dim_to_out_index_tmp[iota_inferred_dim[i_dim]] = (
            iota_inferred_index[i_idx]
        )

    out_index_set = tuple(
        i_inferred_dim_to_out_index_tmp[i] for i in range(len(v_access_out))
    )
    del i_inferred_dim_to_out_index_tmp

    # Step 8, 9.
    i_inferred_pos_to_idx_set_tmp: dict[int, tuple[str, ...]] = {}
    i = next(iter(v_output))
    for j in v_arg_pos:
        i_accesses_in = induced_dag.succs(i) & induced_dag.succs(j) & v_access_in
        i_inferred_dim_to_idx: dict[int, str] = {}
        for i_access_in in i_accesses_in:
            (i_dim,) = induced_dag.preds(i_access_in) & v_dims
            (i_idx,) = induced_dag.preds(i_access_in) & v_index
            i_inferred_dim_to_idx[iota_inferred_dim[i_dim]] = iota_inferred_index[
                i_idx
            ]

        i_inferred_pos_to_idx_set_tmp[iota_inferred_arg_pos[j]] = tuple(
            i_inferred_dim_to_idx[i_dim] for i_dim in range(len(i_accesses_in))
        )

    in_index_sets = tuple(
        i_inferred_pos_to_idx_set_tmp[i] for i in range(len(v_arg_pos))
    )
    del i_inferred_pos_to_idx_set_tmp

    # Step 10.
    inferred_resultxpos_to_arg_tmp: dict[tuple[int, int], str] = {}

    for i in v_output:
        for j in v_arg_pos:
            i_access_ins = induced_dag.succs(i) & induced_dag.succs(j)
            (i_arg,) = reduce(
                frozenset.union,
                [
                    induced_dag.succs(i_access_in) & v_arg
                    for i_access_in in i_access_ins
                ],
                cast("frozenset[int]", frozenset()),
            )
            inferred_resultxpos_to_arg_tmp[
                iota_inferred_output[i], iota_inferred_arg_pos[j]
            ] = iota_inferred_arg[i_arg]

    arg_names = tuple(
        tuple(inferred_resultxpos_to_arg_tmp[i, j] for j in range(len(v_arg_pos)))
        for i in range(len(v_output))
    )
    del inferred_resultxpos_to_arg_tmp

    index_to_length_pred = Map(
        (j, i) for i in v_length for j in induced_dag.succs(i) & v_index
    )
    ilength_to_size_param_name = Map(
        (i, iota_inferred_index[j].upper())
        for i in v_length
        if i not in induced_dag.iota_length
        for j in induced_dag.succs(i) & v_index
    )

    inferred_index_to_length: Map[str, ShapeComponentT] = Map(
        (
            iota_inferred_index[i],
            induced_dag.iota_length.get(
                index_to_length_pred[i],
                SizeParam(iota_inferred_index[i].upper()),
            ),
        )
        for i in v_index
    )

    inferred_arg_to_dtype_tmp: dict[str, np.dtype[Any]] = {}
    for i in v_arg:
        (i_dtype,) = induced_dag.preds(i) & v_dtype
        inferred_arg_to_dtype_tmp[iota_inferred_arg[i]] = induced_dag.iota_dtype[
            i_dtype
        ]
    inferred_arg_to_dtype = Map(inferred_arg_to_dtype_tmp)
    del inferred_arg_to_dtype_tmp

    # Step 11.
    from feinsum.make_einsum import array, batched_einsum

    return batched_einsum(
        f"{','.join("".join(idx_set) for idx_set in in_index_sets)} -> "
        f"{''.join(out_index_set)}",
        [
            [
                array(
                    arg_name,
                    [inferred_index_to_length[idx] for idx in idx_set],
                    inferred_arg_to_dtype[arg_name],
                )
                for idx_set, arg_name in zip(in_index_sets, arg_row, strict=True)
            ]
            for arg_row in arg_names
        ],
    ), Map(
        {
            **iota_inferred_index,
            **iota_inferred_arg,
            **ilength_to_size_param_name,
            **{
                i: "_fe_out" if i_output < 1 else f"_fe_out_{i_output - 1}"
                for i, i_output in iota_inferred_output.items()
            },
        }
    )


def _get_canonicalized_einsum_with_subst_mapping(
    einsum: BatchedEinsum,
) -> tuple[BatchedEinsum, frozenbidict[str, str]]:
    """
    Returns a tuple of the form ``(canonicalized_einsum, subst_map)`` where
    *canonicalized_einsum* is an instance of :class:`BatchedEinsum` which is the
    canonicalized version of *einsum* and *subst_map* is the mapping from entities
    of *einsum* to the variables in `*canonicalized_einsum*.
    """
    induced_dag, iota_src = to_induced_dag(einsum)
    induced_dag, perm = induced_dag.to_bliss_canonicalized_induced_dag()
    einsum, iota_dst = from_induced_dag(induced_dag)
    assert len(iota_src) == len(iota_dst)
    return einsum, frozenbidict(
        (name_in_src, iota_dst[perm[i]]) for i, name_in_src in iota_src.items()
    )


def canonicalize_einsum(einsum: BatchedEinsum) -> BatchedEinsum:
    """
    Returns a canonicalized form of *einsum*.

    .. note::

        - Refer to (TODO PAPER) for a definition of isomorphism among
          fused einsums.
    """
    return _get_canonicalized_einsum_with_subst_mapping(einsum)[0]


def get_substitution_mapping_between_isomorphic_batched_einsums(
    batched_einsum_from: BatchedEinsum, batched_einsum_to: BatchedEinsum
) -> Mapping[str, str]:
    """
    Returns the isomorphism mapping from the entities of *batched_einsum_from* to
    the entities of *batched_einsum_to*.

    .. note::

        Raises a :class:`ValueError` if the two batched einsums are not isomorphic.
    """
    canon_batched_einsum_from, subst_map_from = (
        _get_canonicalized_einsum_with_subst_mapping(batched_einsum_from)
    )
    canon_batched_einsum_to, subst_map_to = (
        _get_canonicalized_einsum_with_subst_mapping(batched_einsum_to)
    )

    if canon_batched_einsum_from != canon_batched_einsum_to:
        raise ValueError("Einsums are not isomorphic.")

    return Map(
        {
            var_in_from_einsum: subst_map_to.inv[var_in_canon_einsum]
            for var_in_from_einsum, var_in_canon_einsum in subst_map_from.items()
        }
    )


# vim: foldmethod=marker



import logging
import os

import loopy as lp
import numpy as np
import pyopencl as cl
from loopy.match import Tagged

import feinsum as f

logging.basicConfig(level="INFO")
logger = logging.getLogger(__name__)


def main(cl_ctx):
    t_unit = lp.make_kernel(
        [
            "{[iel_0, idof_0, jdof_0, r_0, x_0]:"
            " 0<=iel_0<1000 and 0<=idof_0,jdof_0<35 and 0<=x_0,r_0<3}",
            "{[iel_1, idof_1, jdof_1, r_1, x_1]:"
            " 0<=iel_1<1000 and 0<=idof_1,jdof_1<35 and 0<=r_1,x_1<3}",
            "{[iel_2, idof_2, ifacedof, iface]:"
            " 0<=iel_2<1000 and 0<=idof_2<35 and 0<=ifacedof<15 and 0<=iface<4}",
        ],
        """
        v_subst(_0, _1, _2) := v[_0, _1, _2]
        u_subst(_0, _1) := u[_0, _1]
        f_0_subst(_0, _1, _2) := F_0[_0, _1, _2]
        f_1_subst(_0, _1, _2) := F_1[_0, _1, _2]
        f_2_subst(_0, _1, _2) := F_2[_0, _1, _2]
        f_3_subst(_0, _1, _2) := F_3[_0, _1, _2]
        jac_subst(_0, _1, _2) := J[_0, _1, _2]
        D_subst(_0, _1, _2) := D[_0, _1, _2]
        jac_face_subst(_0, _1) := Jface[_0, _1]
        L_subst(_0, _1, _2) := L[_0, _1, _2]

        # ----- Div(v)
        with {tags=div}
            div_out[iel_0,idof_0] = sum([jdof_0,r_0,x_0], \
                                        jac_subst(x_0, r_0, iel_0)*D_subst(r_0,idof_0,jdof_0)*v_subst(x_0, iel_0,jdof_0))
        end

        ... gbarrier {id=g_barrier_0, dep_query=(writes:div_out)}
        # ----- Grad(u)
        with {dep=g_barrier_0, tags=grad}
            grad_out[x_1, iel_1, idof_1] = sum([jdof_1, r_1], \
                                               jac_subst(x_1, r_1, iel_1)*D_subst(r_1, idof_1, jdof_1)*u_subst(iel_1, jdof_1))
        end

        ... gbarrier {id=g_barrier_1, dep_query=(writes:grad_out)}
        # ----- Lift(f*)
        with {dep=g_barrier_1, tags=lift}
            lift_0[iel_2, idof_2] = sum([iface, ifacedof], \
                                        L_subst(idof_2, iface, ifacedof)*jac_face_subst(iface, iel_2)*f_0_subst(iface, iel_2, ifacedof))
            lift_1[iel_2, idof_2] = sum([iface, ifacedof], \
                                        L_subst(idof_2, iface, ifacedof)*jac_face_subst(iface, iel_2)*f_1_subst(iface, iel_2, ifacedof))
            lift_2[iel_2, idof_2] = sum([iface, ifacedof], \
                                        L_subst(idof_2, iface, ifacedof)*jac_face_subst(iface, iel_2)*f_2_subst(iface, iel_2, ifacedof))
            lift_3[iel_2, idof_2] = sum([iface, ifacedof], \
                                        L_subst(idof_2, iface, ifacedof)*jac_face_subst(iface, iel_2)*f_3_subst(iface, iel_2, ifacedof))
        end
        """,  # noqa: E501
    )

    t_unit = lp.add_dtypes(
        t_unit,
        {
            arg.name: np.float64
            for arg in t_unit.default_entrypoint.args
            if arg.is_input
        },
    )
    ref_t_unit = t_unit.copy()

    # {{{ feinsum transformations

    grad_einsum, _ = f.get_a_matched_einsum(t_unit, insn_match=Tagged("grad"))
    div_einsum, _ = f.get_a_matched_einsum(t_unit, insn_match=Tagged("div"))
    lift_einsum, _ = f.get_a_matched_einsum(t_unit, insn_match=Tagged("lift"))

    # {{{ auto-tune

    from feinsum.tuning.impls import (
        ifj_fe_fej_to_ei,
        xre_rij_ej_to_xei,
        xre_rij_xej_to_ei,
    )

    f.autotune(
        grad_einsum,
        os.path.abspath(xre_rij_ej_to_xei.__file__),
        cl_ctx,
        long_dim_length=1000,
        stop_after=3,
    )
    f.autotune(
        div_einsum,
        os.path.abspath(xre_rij_xej_to_ei.__file__),
        cl_ctx,
        long_dim_length=1000,
        stop_after=3,
    )
    f.autotune(
        lift_einsum,
        os.path.abspath(ifj_fe_fej_to_ei.__file__),
        cl_ctx,
        long_dim_length=1000,
        stop_after=3,
    )

    # }}}

    fast_grad_einsum = max(
        f.query(grad_einsum, cl_ctx), key=lambda q: q.giga_op_rate(np.float64)
    )
    fast_div_einsum = max(
        f.query(div_einsum, cl_ctx), key=lambda q: q.giga_op_rate(np.float64)
    )
    fast_lift_einsum = max(
        f.query(lift_einsum, cl_ctx), key=lambda q: q.giga_op_rate(np.float64)
    )

    t_unit = fast_grad_einsum.transform(t_unit, insn_match=Tagged("grad"))
    t_unit = fast_lift_einsum.transform(t_unit, insn_match=Tagged("lift"))
    t_unit = fast_div_einsum.transform(t_unit, insn_match=Tagged("div"))

    # }}}

    lp.auto_test_vs_ref(ref_t_unit, cl_ctx, t_unit)


if __name__ == "__main__":
    cl_ctx = cl.create_some_context()
    main(cl_ctx)



import json
import sqlite3 as sql

import numpy as np

import feinsum as f


def record_into_db(
    cursor: sql.Cursor,
    *,
    einsum: f.BatchedEinsum,
    transform_id: str,
    transform_params: str,
    runtime_in_sec: float,
    compiler_version: str,
    giga_op_info: str,
    timestamp: str,
):
    from feinsum.sql_utils import (
        dump_index_to_length,
        dump_use_matrix,
        dump_value_to_dtype,
    )

    einsum = f.canonicalize_einsum(einsum)
    subscripts = einsum.get_subscripts()
    index_to_length = dump_index_to_length(einsum)
    use_matrix = dump_use_matrix(einsum)
    value_to_dtype = dump_value_to_dtype(einsum)
    cursor.execute(
        "INSERT INTO FEINSUM_TIMING_FACTS"
        " (subscripts, index_to_length, use_matrix,"
        "  value_to_dtype, device_name, transform_id,"
        "  transform_params, runtime_in_sec,"
        "  compiler_version, giga_op_info, timestamp)"
        " VALUES (?,?,?,?,?,?,?,?,?,?,?)",
        (
            subscripts,
            index_to_length,
            use_matrix,
            value_to_dtype,
            "NVIDIA_TITAN_V",
            transform_id,
            transform_params,
            runtime_in_sec,
            compiler_version,
            giga_op_info,
            timestamp,
        ),
    )


def main():
    conn = sql.connect(
        "file:/home/kgk2/projects/ceesd/feinsum"
        "/data/transform_archive_v2.sqlite?mode=ro",
        uri=True,
    )
    cursor = conn.cursor()
    cursor.execute(
        "SELECT "
        " subscripts,"
        " index_to_length,"
        " use_matrix,"
        " value_to_dtype,"
        " transform_id,"
        " transform_params,"
        " runtime_in_sec,"
        " compiler_version,"
        " giga_op_info,"
        " timestamp"
        " FROM NVIDIA_TITAN_V;"
    )
    all_timing_facts = cursor.fetchall()
    conn.close()

    conn = sql.connect("transform_archive_v3.sqlite")
    cursor = conn.cursor()
    cursor.execute(
        "CREATE TABLE FEINSUM_TIMING_FACTS ("
        " ID INTEGER PRIMARY KEY AUTOINCREMENT,"
        " subscripts TEXT,"
        " index_to_length TEXT,"
        " use_matrix TEXT,"
        " value_to_dtype TEXT,"
        " device_name TEXT,"
        " transform_id TEXT,"
        " transform_params TEXT,"
        " runtime_in_sec REAL,"
        " compiler_version TEXT,"
        " giga_op_info TEXT,"
        " timestamp TEXT"
        ")"
    )

    print(len(all_timing_facts))
    for fact in all_timing_facts:
        (
            subscripts,
            idx_to_len,
            use_matrix,
            val_to_dtype,
            transform_id,
            transform_params,
            runtime_in_sec,
            compiler_version,
            giga_op_info,
            timestamp,
        ) = fact
        input_subscripts = subscripts.split("->")[0]
        idx_to_len = json.loads(idx_to_len)
        val_to_dtype = {
            val: np.dtype(dtype) for val, dtype in json.loads(val_to_dtype).items()
        }
        use_matrix = [
            [frozenset(uses) for uses in use_row]
            for use_row in json.loads(use_matrix)
        ]
        arg_shapes = [
            [idx_to_len.get(idx, np.inf) for idx in input_subscript]
            for input_subscript in input_subscripts.split(",")
        ]
        einsum = f.batched_einsum(
            subscripts, arg_shapes, use_matrix, value_to_dtype=val_to_dtype
        )
        record_into_db(
            cursor,
            einsum=einsum,
            transform_id=transform_id,
            transform_params=transform_params,
            runtime_in_sec=runtime_in_sec,
            compiler_version=compiler_version,
            giga_op_info=giga_op_info,
            timestamp=timestamp,
        )

    conn.commit()
    conn.close()


if __name__ == "__main__":
    main()



import logging

import loopy as lp
import numpy as np
import pyopencl as cl

import feinsum as f

logging.basicConfig(level="INFO")
logger = logging.getLogger(__name__)


def get_face_mass_einsum(nface, nfacedofs, nvoldofs):
    return f.batched_einsum(
        "ef, fij, fej -> ei",
        [(np.inf, nface), (nface, nvoldofs, nfacedofs), (nface, np.inf, nfacedofs)],
        dtypes="float64",
        use_matrix=[
            [{"J"}, {"R"}, {"v0"}],
            [{"J"}, {"R"}, {"v1"}],
            [{"J"}, {"R"}, {"v2"}],
            [{"J"}, {"R"}, {"v3"}],
        ],
    )


def variant_0(t_unit):
    """
    Un-fuse all the loops.
    """
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "f"],
        within="id:insn",
        new_inames=["e_0", "i_0", "j_0", "f_0"],
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "f"],
        within="id:insn_0",
        new_inames=["e_1", "i_1", "j_1", "f_1"],
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "f"],
        within="id:insn_1",
        new_inames=["e_2", "i_2", "j_2", "f_2"],
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "f"],
        within="id:insn_2",
        new_inames=["e_3", "i_3", "j_3", "f_3"],
    )

    for iel, idof in [
        ("e_0", "i_0"),
        ("e_1", "i_1"),
        ("e_2", "i_2"),
        ("e_3", "i_3"),
    ]:
        t_unit = lp.split_iname(t_unit, iel, 4, outer_tag="g.0", inner_tag="l.1")
        t_unit = lp.split_iname(t_unit, idof, 8, outer_tag="unr", inner_tag="l.0")

    return t_unit


def variant_1(t_unit):
    """
    Simple work division strategy.
    """
    ref_einsum = get_face_mass_einsum(nface=4, nvoldofs=35, nfacedofs=15)

    subst_map = f.match_t_unit_to_einsum(t_unit, ref_einsum)
    iface = subst_map["f"]
    e = subst_map["e"]
    idof = subst_map["i"]

    t_unit = lp.tag_inames(t_unit, {iface: "unr"})
    t_unit = lp.split_iname(t_unit, e, 8, outer_tag="g.0", inner_tag="l.1")
    t_unit = lp.split_iname(t_unit, idof, 4, inner_tag="l.0", outer_tag="ilp")

    return t_unit


def variant_2(t_unit, insn_match=None, kernel_name=None):
    from pymbolic import variables

    kernel_name = kernel_name or t_unit.default_entrypoint.name

    ncells_per_group = 16
    nworkitems_per_cell = 12

    ref_einsum = get_face_mass_einsum(nface=4, nvoldofs=35, nfacedofs=15)

    subst_map = f.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    e = subst_map["e"]
    iface = subst_map["f"]
    vs = [subst_map[f"v{i}"] for i in range(4)]
    idof = subst_map["i"]
    j = subst_map["j"]
    J = subst_map["J"]
    R = subst_map["R"]

    knl = t_unit[kernel_name]
    output_names = ["_fe_out"] + [f"_fe_out_{i}" for i in range(4)]

    for i in range(4):
        knl = lp.extract_multiplicative_terms_in_sum_reduction_as_subst(
            knl,
            within=f"writes:{output_names[i]}",
            subst_name=f"subst{i}",
            arguments=variables(f"{iface} {e} {j}"),
            terms_filter=lambda x: {J, vs[i]} & f.get_call_ids(x),
        )

    t_unit = t_unit.with_kernel(knl)

    t_unit = lp.split_iname(
        t_unit,
        e,
        ncells_per_group,
        inner_iname="e_inner",
        outer_iname="e_outer",
        outer_tag="g.0",
        inner_tag="l.1",
    )

    # {{{ fetch 'R'

    t_unit = lp.precompute(
        t_unit,
        R,
        [iface, idof, j],
        precompute_outer_inames=frozenset(["e_outer"]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        default_tag=None,
        precompute_inames=["f_Rprftch", "i_Rprftch", "j_Rprftch"],
    )

    t_unit = lp.split_iname(
        t_unit, "i_Rprftch", ncells_per_group, inner_tag="l.1", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, "j_Rprftch", nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    logger.info("Prefetched 'R'")

    # }}}

    t_unit = lp.rename_iname(
        t_unit, idof, "i_0", within="writes:_fe_out or writes:_fe_out_0"
    )
    t_unit = lp.rename_iname(
        t_unit, iface, "f_0", within="writes:_fe_out or writes:_fe_out_0"
    )
    t_unit = lp.rename_iname(
        t_unit, j, "j_0", within="writes:_fe_out or writes:_fe_out_0"
    )

    t_unit = lp.rename_iname(
        t_unit, idof, "i_1", within="writes:_fe_out_1 or writes:_fe_out_2"
    )
    t_unit = lp.rename_iname(
        t_unit, iface, "f_1", within="writes:_fe_out_1 or writes:_fe_out_2"
    )
    t_unit = lp.rename_iname(
        t_unit, j, "j_1", within="writes:_fe_out_1 or writes:_fe_out_2"
    )

    t_unit = lp.split_iname(
        t_unit, "i_0", nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, "i_1", nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    for igrp, prcmpt_tmp, subst in [
        (0, "prcmpt_tmp_0", "subst0"),
        (0, "prcmpt_tmp_1", "subst1"),
        (1, "prcmpt_tmp_0", "subst2"),
        (1, "prcmpt_tmp_1", "subst3"),
    ]:
        logger.info(f"Precomputing {subst}")
        t_unit = lp.precompute(
            t_unit,
            subst,
            sweep_inames=[f"f_{igrp}", "e_inner", f"j_{igrp}"],
            precompute_outer_inames=frozenset(["e_outer"]),
            precompute_inames=[
                f"f_prftch_{igrp}",
                f"e_prftch_{igrp}",
                f"j_prftch_{igrp}",
            ],
            default_tag=None,
            temporary_name=prcmpt_tmp,
            temporary_address_space=lp.AddressSpace.LOCAL,
        )

    for igrp in [0, 1]:
        t_unit = lp.split_iname(
            t_unit,
            f"e_prftch_{igrp}",
            ncells_per_group,
            inner_tag="l.1",
            outer_tag="unr",
        )
        t_unit = lp.split_iname(
            t_unit,
            f"j_prftch_{igrp}",
            nworkitems_per_cell,
            inner_tag="l.0",
            outer_tag="unr",
        )

    t_unit = lp.add_dependency(
        t_unit, "id:subst2 or id:subst3", "writes:_fe_out or writes:_fe_out_0"
    )

    logger.info("Done with transformations.")
    return t_unit


def main():
    cl_ctx = cl.create_some_context()

    expr = get_face_mass_einsum(nface=4, nvoldofs=35, nfacedofs=15)
    print(
        f.stringify_comparison_vs_roofline(
            expr,
            cl_ctx=cl_ctx,
            transform=variant_2,
            long_dim_length=1000,
            ignore_unknown_device=True,  # For CI
        )
    )


if __name__ == "__main__":
    main()



import logging

import loopy as lp
import numpy as np
import pyopencl as cl

import feinsum as f

logger = logging.getLogger(__name__)


def get_div_einsum(ndofs, ndim):
    return f.batched_einsum(
        "es, sij, ej -> ei",
        [(np.inf, ndim), (ndim, ndofs, ndofs), (np.inf, ndofs)],
        dtypes="float64",
        use_matrix=[
            [{"Jx"}, {"R"}, {"ux"}],
            [{"Jy"}, {"R"}, {"uy"}],
            [{"Jz"}, {"R"}, {"uz"}],
        ],
    )


def variant_0(t_unit):
    """
    Un-fuse all the loops.
    """
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "s"],
        within="id:insn",
        new_inames=["e_0", "i_0", "j_0", "s_0"],
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "s"],
        within="id:insn_0",
        new_inames=["e_1", "i_1", "j_1", "s_1"],
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["e", "i", "j", "s"],
        within="id:insn_1",
        new_inames=["e_2", "i_2", "j_2", "s_2"],
    )

    for iel, _ in [("e_0", "i_0"), ("e_1", "i_1"), ("e_2", "i_2")]:
        t_unit = lp.split_iname(t_unit, iel, 32, outer_tag="g.0", inner_tag="l.0")

    return t_unit


def variant_1(t_unit, insn_match=None, kernel_name=None):
    """
    Simple work division strategy.
    """
    t_unit = lp.tag_inames(t_unit, {"s": "unr"})
    t_unit = lp.split_iname(t_unit, "e", 8, outer_tag="g.0", inner_tag="l.1")
    t_unit = lp.split_iname(t_unit, "i", 4, inner_tag="l.0", outer_tag="ilp")

    return t_unit


def variant_3(t_unit):
    # Based on https://github.com/nchristensen/grudge/'s transformation
    # algorithm
    assert t_unit.default_entrypoint.arg_dict["R"].shape == (3, 35, 35)
    imatrix = "s"
    iel = "e"
    idof = "i"

    dof_vecs = [
        arg.name
        for arg in t_unit.default_entrypoint.args
        if arg.name in ["ux", "uy", "uz"]
    ]
    t_unit = lp.tag_inames(t_unit, {imatrix: "unr"})
    t_unit = lp.split_iname(t_unit, iel, 16, outer_tag="g.0")
    t_unit = lp.split_iname(
        t_unit, f"{iel}_inner", 8, outer_tag="ilp", inner_tag="l.0"
    )
    t_unit = lp.tag_inames(t_unit, {idof: "l.1"})
    for vec in dof_vecs:
        t_unit = lp.add_prefetch(
            t_unit,
            vec,
            f"j,{iel}_inner_outer,{iel}_inner_inner",
            temporary_name=f"{vec}_prftch",
            default_tag="l.auto",
        )
    t_unit = lp.add_inames_for_unused_hw_axes(t_unit)
    return t_unit


def main():
    cl_ctx = cl.create_some_context()

    expr = get_div_einsum(ndofs=35, ndim=3)
    print(
        f.stringify_comparison_vs_roofline(
            expr,
            cl_ctx=cl_ctx,
            transform=variant_1,
            long_dim_length=1000,
            ignore_unknown_device=True,  # For CI
        )
    )


if __name__ == "__main__":
    main()



from typing import Any

import loopy as lp
import loopy.match as lp_match

import feinsum as fnsm
import feinsum.loopy_utils as lp_utils
from feinsum.tuning import IntParameter


@fnsm.tuning.einsum_arg("noutputs", lambda e: e.b)
@fnsm.tuning.einsum_arg("ndofs", lambda e: e.shape[1])
@fnsm.tuning.transform_param("nworkitems_per_e", lambda e: IntParameter(8, 8))
@fnsm.tuning.transform_param("n_e_per_wg", lambda e: IntParameter(4, 4))
def transform(
    t_unit: lp.TranslationUnit,
    ndofs: int,
    noutputs: int,
    n_e_per_wg: int,
    nworkitems_per_e: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    if n_e_per_wg * nworkitems_per_e > 600:
        raise fnsm.InvalidParameterError("Block dimension limit exceeded")

    kernel_name = kernel_name or t_unit.default_entrypoint.name
    within = lp_match.parse_match(insn_match)

    ref_einsum = fnsm.batched_einsum(
        "e,ij,ej->ei",
        [
            [
                fnsm.array("J", "Nel"),
                fnsm.array("D", (ndofs, ndofs)),
                fnsm.array(f"u{i}", ("Nel", ndofs)),
            ]
            for i in range(noutputs)
        ],
    )
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, kernel_name=kernel_name, insn_match=within
    )
    vng = t_unit[kernel_name].get_var_name_generator()

    e = subst_map["e"]
    i = subst_map["i"]
    j = subst_map["j"]
    e_inner, e_outer = vng(f"{e}_inner"), vng(f"{e}_outer")
    i_inner, i_outer = vng(f"{i}_inner"), vng(f"{i}_outer")
    J = subst_map["J"]
    ref_outs = ["_fe_out"] + [f"_fe_out_{iout}" for iout in range(noutputs - 1)]
    outs = tuple(subst_map[ref_out] for ref_out in ref_outs)
    del ref_outs

    t_unit = t_unit.with_kernel(
        lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(
            t_unit[kernel_name], j
        )
    )

    t_unit = lp.split_iname(t_unit, e, n_e_per_wg, inner_tag="l.1", outer_tag="g.0")
    t_unit = lp.split_iname(
        t_unit, i, nworkitems_per_e, inner_tag="l.0", outer_tag="unr"
    )

    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        J,
        sweep_inames=[],
        precompute_outer_inames=frozenset([e_outer, e_inner, i_inner]),
        default_tag="unr",
        within=within,
        temporary_address_space=lp.AddressSpace.PRIVATE,
    )

    for out in outs[1:]:
        t_unit = lp.duplicate_inames(
            t_unit,
            (i_outer, j),
            within=lp_match.And((within, lp_match.Writes(out))),
            tags={
                i_outer: t_unit[kernel_name].inames[i_outer].tags,
                j: t_unit[kernel_name].inames[j].tags,
            },
        )
    return t_unit


if __name__ == "__main__":
    import os

    import pyopencl as cl

    Ndof = 4
    Nfields = 16

    cl_ctx = cl.create_some_context()

    expr = fnsm.batched_einsum(
        "e,ij,ej->ei",
        [
            [
                fnsm.array("J", "Nel"),
                fnsm.array("D", (Ndof, Ndof)),
                fnsm.array(f"u{i}", ("Nel", Ndof)),
            ]
            for i in range(Nfields)
        ],
    )

    fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)

# vim: fdm=marker



import logging

import loopy as lp

import feinsum as fnsm
import feinsum.loopy_utils as lp_utils

logging.basicConfig(level="INFO")
logger = logging.getLogger(__name__)


def transform_3d_p4_grad(t_unit, insn_match=None, kernel_name=None):
    """
    A transformation that performs ~1.8 TFlOps/s on a TitanV.
    """
    from loopy.match import parse_match
    from pymbolic import variables

    insn_match = parse_match(insn_match)

    # {{{ define ref_einsum; get subst_map

    ndim = 3
    ndofs = 35

    ref_einsum = fnsm.einsum(
        "xer,rij,ej->xei",
        fnsm.array(
            "J",
            (
                ndim,
                "E",
                ndim,
            ),
        ),
        fnsm.array("R", (ndim, ndofs, ndofs)),
        fnsm.array("u", ("E", ndofs)),
    )

    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match
    )

    # }}}

    # {{{ transform parameters

    ncells_per_workgroup = 9
    nworkitems_per_cell = 7
    j_tile_len = 9
    i_tile_len = 35

    # }}}

    # {{{ read variables

    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    R = subst_map["R"]
    J = subst_map["J"]
    u = subst_map["u"]
    x = subst_map["x"]
    r = subst_map["r"]
    out = subst_map["_fe_out"]
    e_inner, e_outer = f"{e}_inner", f"{e}_outer"
    i_inner, i_outer = f"{i}_inner", f"{i}_outer"
    j_inner = f"{j}_inner"
    i_tile, j_tile = f"{i}_tile", f"{j}_tile"
    J_prftch = f"{J}_prftch"
    e_prftch = f"{e}_prftch"
    j_prftch = f"{j}_prftch"
    i_prftch = f"{i}_prftch"
    r_prftch = f"{r}_prftch"
    r_prcmpt = f"{r}_prcmpt"

    # }}}

    # {{{ term hoisting to match the flop count of opt_einsum

    from loopy.symbolic import Reduction

    knl = t_unit.default_entrypoint

    knl = lp.split_reduction_inward(knl, j)
    knl = lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, j)
    knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        within=None,
        subst_name="subst",
        arguments=variables(f"{r} {i} {e}"),
        terms_filter=lambda x: isinstance(x, Reduction),
    )

    t_unit = t_unit.with_kernel(knl)

    # }}}

    t_unit = lp.split_iname(t_unit, i, i_tile_len, outer_iname=i_tile)
    t_unit = lp.split_iname(t_unit, j, j_tile_len, outer_iname=j_tile)

    t_unit = lp.rename_iname(t_unit, i_inner, i)
    t_unit = lp.rename_iname(t_unit, j_inner, j)
    t_unit = lp.split_iname(
        t_unit, e, ncells_per_workgroup, outer_tag="g.0", inner_tag="l.1"
    )
    t_unit = lp.split_iname(t_unit, i, nworkitems_per_cell, inner_tag="l.0")

    t_unit = lp.precompute(
        t_unit,
        J,
        sweep_inames=[r, x],
        precompute_outer_inames=frozenset([e_inner, e_outer, i_inner]),
        temporary_address_space=lp.AddressSpace.PRIVATE,
        temporary_name=J_prftch,
    )

    # {{{ TODO: Make precompute smarter (should be a single precompute call)

    t_unit = lp.precompute(
        t_unit,
        "subst",
        sweep_inames=[r],
        precompute_outer_inames=frozenset(
            {e_inner, e_outer, i_inner, i_outer, i_tile}
        ),
        precompute_inames=r_prcmpt,
        temporary_name="tmp_hoist",
        temporary_address_space=lp.AddressSpace.PRIVATE,
        compute_insn_id="insn_hoist",
    )
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, i_outer, only_var_names=["tmp_hoist"]
    )

    t_unit = lp.duplicate_inames(t_unit, i_outer, "id:insn_hoist", "i_outer_hoist")

    # }}}

    # {{{ Move 'u ' to shared.

    # Prefetch 'u' within the tile
    t_unit = lp.precompute(
        t_unit,
        u,
        sweep_inames=[e_inner, j],
        precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        precompute_inames=[e_prftch, j_prftch],
        default_tag=None,
    )

    t_unit = lp.split_iname(
        t_unit, e_prftch, ncells_per_workgroup, inner_tag="l.1", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, j_prftch, nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
    )

    # }}}

    # {{{ Move 'R' to shared.

    t_unit = lp.precompute(
        t_unit,
        R,
        sweep_inames=[r_prcmpt, i_inner, "i_outer_hoist", j],
        precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
        temporary_address_space=lp.AddressSpace.LOCAL,
        precompute_inames=[r_prftch, i_prftch, j_prftch],
        default_tag=None,
        within="id:insn_hoist",
    )

    if 0:
        # This branch improves perf. by 20% but, something in loopy
        # non-deterministically leads to very ugly domains.
        t_unit = lp.join_inames(t_unit, [r_prftch, i_prftch, j_prftch], "i_Rprftch")

        t_unit = lp.split_iname(
            t_unit,
            "i_Rprftch",
            ncells_per_workgroup * nworkitems_per_cell,
            outer_tag="unr",
        )

        t_unit = lp.split_iname(
            t_unit,
            "i_Rprftch_inner",
            nworkitems_per_cell,
            inner_tag="l.0",
            outer_tag="l.1",
        )
    else:
        t_unit = lp.split_iname(
            t_unit, i_prftch, ncells_per_workgroup, inner_tag="l.1", outer_tag="unr"
        )
        t_unit = lp.split_iname(
            t_unit, j_prftch, nworkitems_per_cell, inner_tag="l.0", outer_tag="unr"
        )

    # }}}

    # {{{ make buffer array smarter (should be a single call to buffer_array)

    t_unit = lp.buffer_array(
        t_unit,
        out,
        buffer_inames=[x],
        init_expression="0",
        default_tag=None,
        temporary_scope=lp.AddressSpace.PRIVATE,
    )
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, i_outer, only_var_names={f"{out}_buf"}
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames=[i_outer],
        within=f"id:init_{out}",
        new_inames=[f"{out}_init_1"],
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames=[i_outer],
        within=f"id:store_{out}",
        new_inames=[f"{out}_store_1"],
    )

    # }}}

    # {{{ must be smarter way of doing this in loopy

    t_unit = lp.realize_reduction(t_unit, insn_id_filter="insn_hoist")
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit,
        frozenset([r_prcmpt, "i_outer_hoist"]),
        only_var_names={f"acc_{j_tile}_{j}"},
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        ["i_outer_hoist", r_prcmpt],
        within=f"id:insn_hoist_{j_tile}_{j}_init",
        new_inames=["i_outer_hoist_init", f"{r_prcmpt}_init"],
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        ["i_outer_hoist", r_prcmpt],
        within="id:insn_hoist",
        new_inames=["i_outer_hoist_store", f"{r_prcmpt}_store"],
    )

    # }}}

    return t_unit
